/*
    Copyright 2008 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include "show_dump.h"

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER             "0.1"
#define BUFFSZ          0xffff
#define PORT            53

/* Length of DNS header. */
#define DNS_HDRLEN      12

/* type values  */
#define T_A             1           /* host address */
#define T_NS            2           /* authoritative name server */
#define T_MD            3           /* mail destination (obsolete) */
#define T_MF            4           /* mail forwarder (obsolete) */
#define T_CNAME         5           /* canonical name */
#define T_SOA           6           /* start of authority zone */
#define T_MB            7           /* mailbox domain name (experimental) */
#define T_MG            8           /* mail group member (experimental) */
#define T_MR            9           /* mail rename domain name (experimental) */
#define T_NULL          10          /* null RR (experimental) */
#define T_WKS           11          /* well known service */
#define T_PTR           12          /* domain name pointer */
#define T_HINFO         13          /* host information */
#define T_MINFO         14          /* mailbox or mail list information */
#define T_MX            15          /* mail routing information */
#define T_TXT           16          /* text strings */
#define T_RP            17          /* responsible person (RFC 1183) */
#define T_AFSDB         18          /* AFS data base location (RFC 1183) */
#define T_X25           19          /* X.25 address (RFC 1183) */
#define T_ISDN          20          /* ISDN address (RFC 1183) */
#define T_RT            21          /* route-through (RFC 1183) */
#define T_NSAP          22          /* OSI NSAP (RFC 1706) */
#define T_NSAP_PTR      23          /* PTR equivalent for OSI NSAP (RFC 1348 - obsolete) */
#define T_SIG           24          /* digital signature (RFC 2535) */
#define T_KEY           25          /* public key (RFC 2535) */
#define T_PX            26          /* pointer to X.400/RFC822 mapping info (RFC 1664) */
#define T_GPOS          27          /* geographical position (RFC 1712) */
#define T_AAAA          28          /* IPv6 address (RFC 1886) */
#define T_LOC           29          /* geographical location (RFC 1876) */
#define T_NXT           30          /* "next" name (RFC 2535) */
#define T_EID           31          /* ??? (Nimrod?) */
#define T_NIMLOC        32          /* ??? (Nimrod?) */
#define T_SRV           33          /* service location (RFC 2052) */
#define T_ATMA          34          /* ??? */
#define T_NAPTR         35          /* naming authority pointer (RFC 2168) */
#define T_KX            36          /* Key Exchange (RFC 2230) */
#define T_CERT          37          /* Certificate (RFC 2538) */
#define T_A6            38          /* IPv6 address with indirection (RFC 2874) */
#define T_DNAME         39          /* Non-terminal DNS name redirection (RFC 2672) */
#define T_OPT           41          /* OPT pseudo-RR (RFC 2671) */
#define T_DS            43          /* Delegation Signature(RFC 3658) */
#define T_IPSECKEY      45          /* draft-ietf-ipseckey-rr */
#define T_RRSIG         46          /* future RFC 2535bis */
#define T_NSEC          47          /* future RFC 2535bis */
#define T_DNSKEY        48          /* future RFC 2535bis */
#define T_TKEY          249         /* Transaction Key (RFC 2930) */
#define T_TSIG          250         /* Transaction Signature (RFC 2845) */
#define T_WINS          65281       /* Microsoft's WINS RR */
#define T_WINS_R        65282       /* Microsoft's WINS-R RR */

/* Class values */
#define C_IN            1           /* the Internet */
#define C_CS            2           /* CSNET (obsolete) */
#define C_CH            3           /* CHAOS */
#define C_HS            4           /* Hesiod */
#define C_NONE          254         /* none */
#define C_ANY           255         /* any */
#define C_FLUSH         (1<<15)     /* High bit is set for MDNS cache flush */



int fast_answer(u8 *ps, int host, int type);
int getxx(u8 *data, u32 *ret, int bits);
int putxx(u8 *data, u32 num, int bits);
u32 resolv(char *host);
void std_err(void);



int     verbose = 0;    // implemented for future updates



int main(int argc, char *argv[]) {
    struct  sockaddr_in peer;
    u32     hostip,
            num;
    int     sd,
            i,
            len,
            on      = 1,
            psz;
    u8      *host   = "127.0.0.1",
            *rbuff,
            *sbuff,
            *lr,
            *pr,
            *ps;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);
    setbuf(stderr, NULL);

    fputs("\n"
        "loDNS "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stderr);

    printf("\n"
        "Usage: %s [host_IP(%s)]\n"
        "\n"
        "use 0.0.0.0 for monitoring the queries without sending replies\n"
        "\n", argv[0], host);

    if(argc > 1) host = argv[1];

    peer.sin_addr.s_addr = resolv(host);
    if(verbose) printf("- IP %s\n", inet_ntoa(peer.sin_addr));
    hostip = ntohl(peer.sin_addr.s_addr);

    peer.sin_addr.s_addr = INADDR_ANY;
    peer.sin_port        = htons(PORT);
    peer.sin_family      = AF_INET;
    psz                  = sizeof(struct sockaddr_in);

    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();

    if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on))
      < 0) std_err();
    if(bind(sd, (struct sockaddr *)&peer, sizeof(struct sockaddr_in))
      < 0) std_err();

    rbuff = malloc(BUFFSZ);
    sbuff = malloc(BUFFSZ);
    if(!rbuff || !sbuff) std_err();

    printf("- start receiving:\n");
    for(;;) {
        len = recvfrom(sd, rbuff, BUFFSZ, 0, (struct sockaddr *)&peer, &psz);
        if(len <= DNS_HDRLEN) continue;
        printf("\n  %s:%hu\n", inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));
                                            if(verbose) show_dump(rbuff, len, stdout);

        pr = rbuff;
        lr = rbuff + len;
        ps = sbuff;

        pr += getxx(pr, &num, 16);          if(verbose) printf("  transaction ID: 0x%04x\n", num);
        ps += putxx(ps, num,  16);          // transaction ID

        pr += getxx(pr, &num, 16);          if(verbose) printf("  flags:          0x%04x\n", num);
        pr += getxx(pr, &num, 16);          if(verbose) printf("  questions:      0x%04x\n", num);
        pr += getxx(pr, &num, 16);          if(verbose) printf("  answer RRs:     0x%04x\n", num);
        pr += getxx(pr, &num, 16);          if(verbose) printf("  authority RRs:  0x%04x\n", num);
        pr += getxx(pr, &num, 16);          if(verbose) printf("  additional RRs: 0x%04x\n", num);

        num = 32768                         // response
            | 0                             // opcode
            | 1024                          // authoritative
            | 0                             // truncated
            | 0                             // recursion desired
            | 0                             // recursion available
            | 0                             // Z
            | 32                            // answer authenticated
            | 0;                            // reply code
        ps += putxx(ps, num,  16);          // flags
        ps += putxx(ps, 1,    16);          // questions
        ps += putxx(ps, 1,    16);          // answer RRs
        ps += putxx(ps, 0,    16);          // authority RRs
        ps += putxx(ps, 0,    16);          // additional RRs

        while(pr < lr) {
                                            if(verbose) printf("  name:         ");
            for(i = 0; ; i++) {
                pr += getxx(pr, &num, 8);
                ps += putxx(ps, num,  8);
                if(!num || (num & 0xc0)) break;
                printf("%s%.*s", i ? "." : "  ", num, pr);
                memcpy(ps, pr, num);
                pr += num;
                ps += num;
            }
            fputc('\n', stdout);
            if(num & 0xc0) break;

            pr += getxx(pr, &num, 16);      if(verbose) printf("  type:           0x%04x\n", num);
            ps += putxx(ps, num,  16);

            pr += getxx(pr, &num, 16);      if(verbose) printf("  class:          0x%04x\n", num);
            ps += putxx(ps, num,  16);
        }

        ps += fast_answer(ps, 0x0c, T_A);
        ps += putxx(ps, 4,      16);        // data length, only IPv4
        ps += putxx(ps, hostip, 32);        // IP

/*
        ps += fast_answer(ps, 0x0c, T_CNAME);
        ps += putxx(ps, 2,      16);
        ps += putxx(ps, 0xc0,   8);         // domain
        ps += putxx(ps, 0x10,   8);

        ps += fast_answer(ps, 0x10, T_MX);
        ps += putxx(ps, 2 + 2,  16);
        ps += putxx(ps, 0,      16);        // preference
        ps += putxx(ps, 0xc0,   8);         // host
        ps += putxx(ps, 0x0c,   8);

        ps += fast_answer(ps, 0x10, T_SOA);
        ps += putxx(ps, 2 + 2 + 20, 16);
        ps += putxx(ps, 0xc0,   8);         // primary name server
        ps += putxx(ps, 0x0c,   8);
        ps += putxx(ps, 0xc0,   8);         // responsible authority mailbox
        ps += putxx(ps, 0x0c,   8);
        ps += putxx(ps, 1,      32);        // serial number
        ps += putxx(ps, 28800,  32);        // refresh interval
        ps += putxx(ps, 7200,   32);        // retry interval
        ps += putxx(ps, 604800, 32);        // expiration limit
        ps += putxx(ps, 86400,  32);        // minimum TTL

        ps += fast_answer(ps, 0x10, T_NS);
        ps += putxx(ps, 2,      16);
        ps += putxx(ps, 0xc0,   8);         // host
        ps += putxx(ps, 0x0c,   8);

        ps += fast_answer(ps, 0x0c, T_PTR);
        ps += putxx(ps, 2,      16);
        ps += putxx(ps, 0xc0,   8);         // host
        ps += putxx(ps, 0x0c,   8);
*/

                                            if(verbose) show_dump(sbuff, ps - sbuff, stdout);

        if(hostip) {                        // if 0 simply monitors without sending data
            sendto(sd, sbuff, ps - sbuff, 0, (struct sockaddr *)&peer, sizeof(struct sockaddr_in));
        }
    }

    close(sd);
    return(0);
}



int fast_answer(u8 *ps, int host, int type) {
    ps += putxx(ps, 0xc0, 8);
    ps += putxx(ps, host, 8);
    ps += putxx(ps, type, 16);
    ps += putxx(ps, C_IN, 16);
    ps += putxx(ps, 1800, 32);
    return(10);
}



int getxx(u8 *data, u32 *ret, int bits) {
    u32     num;
    int     i,
            bytes;

    bytes = bits >> 3;
    for(num = i = 0; i < bytes; i++) {
        num |= (data[i] << ((bytes - i - 1) << 3));
    }
    *ret = num;
    return(bytes);
}



int putxx(u8 *data, u32 num, int bits) {
    int     i,
            bytes;

    bytes = bits >> 3;
    for(i = 0; i < bytes; i++) {
        data[i] = num >> ((bytes - i - 1) << 3);
    }
    return(bytes);
}



u32 resolv(char *host) {
    struct  hostent *hp;
    u32     host_ip;

    host_ip = inet_addr(host);
    if(host_ip == htonl(INADDR_NONE)) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolve hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u32 *)(hp->h_addr);
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif



